from enstools.feature.util.data_utils import pb_str_to_datetime
from enstools.feature.tracking import TrackingStrategy

class DataGraph:
    """
    Abstraction for graphs:
    It contains methods to iterate a graph and to generate the tracks.

    It can be initialized just using a 'track', see the template.

    It can also be initialized using the set_description from protobuf
    and the tracking strategy. Then the graph represents the full structure
    and not only a single track.

    To generate tracks, an instance of the latter one is needed.
    """

    def __init__(self, set_desc=None, tr_tech=None, track=None, fix=False, remove_transitive=False):
        self.set_desc = set_desc

        # sort the edges in the graph by time.
        # makes it easier to traverse later.
        if set_desc is not None:
            self.graph = set_desc.graph
            self.graph.edges.sort(key=lambda item: item.parent.time)
        if track is not None:
            self.graph = track
            self.graph.edges.sort(key=lambda item: item.parent.time)

        self.tr_tech = tr_tech

        if fix:
            self.fix_graph(remove_transitive=remove_transitive)


    def get_earliest_nodes(self):
        """
        Get the nodes which are earliest in the graph.
        Note that multiple earliest nodes might exist.

        Returns
        -------
        List of nodes with earliest timestamp.
        """
        if len(self.graph.edges) == 0:
            return []

        earliest_time = self.graph.edges[0].parent.time
        nodes = [self.graph.edges[0].parent]

        # iterate over first nodes until time changes or end of list
        idx = 1
        while idx < len(self.graph.edges) and self.graph.edges[idx].parent.time == earliest_time:
            nodes.append(self.graph.edges[idx].parent)
            idx += 1

        return nodes

    def get_latest_nodes(self):
        """
        Get the nodes which are latest in the graph.
        Note that multiple latest nodes might exist.

        Returns
        -------
        List of nodes with latest timestamp.
        """
        if len(self.graph.edges) == 0:
            return []

        latest_time = self.graph.edges[-1].parent.time
        nodes = [self.graph.edges[-1].parent]

        # iterate over first nodes until time changes or end of list
        idx = len(self.graph.edges) - 2
        while idx >= 0 and self.graph.edges[idx].parent.time == latest_time:
            nodes.append(self.graph.edges[idx].parent)
            idx -= 1

        return nodes

    def fix_graph(self, remove_transitive=False):
        """
        Fixes graph by removing orphaned edges. (Edges where endnode is not a valid node)

        Returns
        -------
        """
        nodes = [e.parent for e in self.graph.edges]

        for edge in self.graph.edges:
            for child in edge.children[:]:
                if child not in nodes:
                    edge.children.remove(child)

        if remove_transitive:
            non_tr_edges = TrackingStrategy.remove_transitive(self.graph.edges, self.tr_tech.pb_reference)
            del self.graph.edges[:]
            self.graph.edges.extend(non_tr_edges)



    def get_parents(self, node):
        """
        Get the parents of a node (previous timestep)
        Parameters
        ----------
        node: The node to get the parents of

        Returns
        -------
        List of nodes representing the parents. Multiple nodes indicate a merge event.
        """
        parents = []
        for edge in self.graph.edges:
            for child in edge.children:
                if child == node:
                    parents.append(edge.parent)
        return parents

    def get_children(self, node):
        """
        Get the children of a node (next timestep)
        Parameters
        ----------
        node: The node to get the children of

        Returns
        -------
        List of nodes representing the children. Multiple nodes indicate a split event.
        """
        for edge in self.graph.edges:
            if edge.parent == node:
                return edge.children
        return []

    def empty_copy(self):
        copied_set_desc = self.tr_tech.pb_reference.TrackableSet()
        copied_set_desc.CopyFrom(self.set_desc)
        del copied_set_desc.graph
        del copied_set_desc.tracks[:]

        return DataGraph(copied_set_desc, self.tr_tech, fix=False)


    def generate_tracks(self, apply_filter=True):
        """
        After tracking graph has been computed, here, tracks can be computed, which are a disjoint subset of graphs
        of the total graph. It is based on a simple heuristic. The nodes are ordered by time. For each non-classified
        node, all downstream nodes are searched. If any of these nodes is already classified, use the same ID.
        Otherwise give this stream a new id (new track).

        Returns
        -------
        None, Tracks are added to set_desc.tracks
        """

        # sort nodes by time of first node (is in key), as list here
        time_sorted_nodes = list(sorted(self.graph.edges, key=lambda item: item.parent.time))

        id_per_node = [None] * len(time_sorted_nodes)

        cur_id = 0

        # iterate over all time sorted identified connections
        # search temporal downstream tracked troughs and group them using a set id
        for con_idx, oc in enumerate(time_sorted_nodes):

            if id_per_node[con_idx] is not None:  # already part of an object
                continue

            # not part of object -> get (temporal) downstream connections = part of same object (return indices of them)
            downstream_node_indices = DataGraph.get_downstream_node_indices(time_sorted_nodes, con_idx)

            # any of downstream nodes already part of an object?
            connected_id = None
            for ds_node_idx in downstream_node_indices:
                if id_per_node[ds_node_idx] is not None:
                    # if connected_object_id is not None:
                    #     print("Double ID, better resolve") # TODO
                    connected_id = id_per_node[ds_node_idx]

            # if so set all nodes to this found id
            if connected_id is not None:
                for ds_node_idx in downstream_node_indices:
                    id_per_node[ds_node_idx] = connected_id
                continue

            # else new path for all object_nodes
            cur_id_needs_update = False
            for ds_node_idx in downstream_node_indices:
                id_per_node[ds_node_idx] = cur_id
                cur_id_needs_update = True

            if cur_id_needs_update:
                cur_id += 1

        # done, now extract every object by id and put them into subgraphs

        print(len(self.set_desc.tracks))

        for object_id in range(cur_id):
            object_idxs = [i for i in range(len(id_per_node)) if id_per_node[i] == object_id]

            cur_object_nodes = [time_sorted_nodes[i] for i in object_idxs]  # states of this object
            # cur_troughs = cur_troughs.sortbytime # already sorted?

            track = self.tr_tech.pb_reference.ObjectGraph()
            track.edges.extend(cur_object_nodes)

            # make a graph out of it
            track_set_ = self.tr_tech.pb_reference.TrackableSet()
            track_set_.CopyFrom(self.set_desc)
            track_set_.graph.CopyFrom(track)

            track_graph = DataGraph(track_set_, self.tr_tech, fix=True)

            if apply_filter:
                # let user change track, reject it, or split it
                adj_tracking_graph = self.tr_tech.adjust_track(track_graph)
                # if user returns a boolean, handle it as well
                if type(adj_tracking_graph) == bool and track_graph:
                    adj_tracking_graph = track_graph
            else:
                adj_tracking_graph = track_graph

            # did user reject this track?
            if adj_tracking_graph is None or len(adj_tracking_graph.graph.edges) == 0:
                continue

            if isinstance(adj_tracking_graph, list) and len(adj_tracking_graph) > 0:
                if not type(adj_tracking_graph[0]) == DataGraph:
                    print("Invalid type of return in adjust_track. Requires bool or list of DataGraphs.")

                for adj_tr_g  in adj_tracking_graph:
                    self.set_desc.tracks.append(adj_tr_g.graph)

            else: # single
                self.set_desc.tracks.append(adj_tracking_graph.graph)
            # if user didnt write any tracks to tg.set_desc.tracks, take track from graph.
            # TODO needs nicer API for sure. TODO fix in AEWs

        return None

    def add_tracks_from(self, data_graph):
        print("Add track from")
        self.set_desc.tracks.extend(data_graph.set_desc.tracks)

    @staticmethod
    def get_downstream_node_indices(graph_list, start_idx, until_time=None):
        """
        Helper method for the track generation. Searches all downstream node indices. (time-wise)

        Parameters
        ----------
        graph_list:
            List of GraphNodes, sorted by parent time
        start_idx:
            Index of GraphNode to start
        until_time:
            latest time to consider, None if don't set a threshold

        Returns
        -------
        List of indices in graph_list containing the downstream nodes of the given one.
        """

        node_indices = [start_idx]
        obj_node_list = [con_.parent for con_ in graph_list]  # this up

        co = graph_list[start_idx]
        node, connected_nodes = co.parent, co.children

        for c_node in connected_nodes:
            # if until_time is set: if c_node time later than until_time end recursion here
            if until_time is not None and pb_str_to_datetime(c_node.time) > until_time:
                continue

            # get index of connected node in graph
            c_node_idx = obj_node_list.index(c_node)

            # call recursively on this connected node
            c_node_downstream_indices = DataGraph.get_downstream_node_indices(graph_list, c_node_idx,
                                                                              until_time=until_time)
            node_indices.extend(c_node_downstream_indices)

        return list(set(node_indices))
